import asyncio
import logging
from typing import Awaitable, Callable, Iterable

import openai
from jims_core.llms.llm_provider import LLMProvider
from jims_core.thread.schema import CommunicationEvent
from openai.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
)
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class Tool[T: BaseModel]:
    def __init__(
        self, name: str, description: str, args_cls: type[T], fn: Callable[[T], Awaitable[str]] | Callable[[T], str]
    ) -> None:
        self.name = name
        self.description = description
        self.args_cls = args_cls
        self.fn = fn
        self.openai_def = openai.pydantic_function_tool(args_cls, name=name, description=description)

    async def call(self, args_json: str) -> str:
        try:
            fn_args = self.args_cls.model_validate_json(args_json)
        except ValueError:
            return f"Invalid tool args: {args_json}"

        if asyncio.iscoroutinefunction(self.fn):
            result = await self.fn(fn_args)
        else:
            result: str = await asyncio.to_thread(self.fn, fn_args)  # type: ignore

        return result


class LLM:
    def __init__(
        self,
        llm_provider: LLMProvider,
        prompt_templates: dict[str, str],
        logger: logging.Logger | None = None,
    ) -> None:
        self.logger = logger or logging.getLogger(__name__)
        self.llm = llm_provider
        self.prompt_templates = prompt_templates

    # Current
    async def generate_cypher_query_with_tools(
        self,
        data_descr: str,
        messages: Iterable,
        tools: list[Tool],
    ) -> tuple[list[ChatCompletionMessageParam], str]:
        tool_names = [t.name for t in tools]
        msgs = make_cypher_query_with_tools_dialog(data_descr, self.prompt_templates, messages, tool_names=tool_names)
        return await self.create_completion_with_tools(msgs, tools=tools)

    async def create_completion_with_tools(
        self,
        messages: list[ChatCompletionMessageParam],
        tools: Iterable[Tool],
    ) -> tuple[list[ChatCompletionMessageParam], str]:
        messages = messages.copy()
        tool_defs = [tool.openai_def for tool in tools]
        tools_map = {tool.name: tool for tool in tools}

        async def _execute_tool_call(tool_call):
            tool_name = tool_call.function.name
            tool = tools_map.get(tool_name)
            if not tool:
                self.logger.error(f"Tool {tool_name} not found!")
                return tool_call.id, f"Tool {tool_name} not found!"

            self.logger.debug(f"Calling tool {tool_name}")
            try:
                tool_res = await tool.call(tool_call.function.arguments)
            except Exception as e:
                self.logger.exception("Error executing tool %s: %s", tool_name, e)
                tool_res = f"Error executing tool {tool_name}: {e}"

            self.logger.debug("Tool %s (%s) result: %s", tool_name, tool.description, tool_res)
            return tool_call.id, tool_res

        max_iters = 5
        for i in range(max_iters):
            msg, tool_calls = await self.llm.chat_completion_with_tools(
                messages=messages,
                tools=tool_defs,
            )

            messages.append(msg.to_dict())  # type: ignore

            if not tool_calls:
                self.logger.debug("No tool calls found. Exiting tool call loop")
                break

            self.logger.debug(f"Tool call iter {i + 1}/{max_iters}")

            # Execute tool calls in parallel
            results = await asyncio.gather(*[_execute_tool_call(t) for t in tool_calls])

            for tool_call_id, tool_res in results:
                messages.append(
                    ChatCompletionToolMessageParam(role="tool", tool_call_id=tool_call_id, content=tool_res)
                )

            if i == max_iters - 1:
                self.logger.warning(f"Reached tool call iteration limit ({max_iters}). Exiting tool call loop")
                finalize_prompt = self.prompt_templates.get("finalize_answer_tmplt", finalize_answer_tmplt)
                finalize_msg = {"role": "system", "content": finalize_prompt}
                final_msg = await self.llm.chat_completion_plain(messages + [finalize_msg])
                messages.append(final_msg.to_dict())  # type: ignore
                break

        for last_msg in reversed(messages):  # sometimes message with final answer is not the last one
            if last_msg.get("role", "") == "assistant" and last_msg.get("content"):
                return messages, str(last_msg.get("content"))
        return messages, ""

    async def generate_no_answer(
        self,
        dialog: list[CommunicationEvent] | None = None,
    ) -> str:
        prompt = self.prompt_templates.get("generate_no_answer_tmplt", generate_no_answer_tmplt)
        messages = [
            {"role": "system", "content": prompt},
            *(dialog or []),
        ]
        response = await self.llm.chat_completion_plain(messages)
        human_answer = "" if response.content is None else response.content.strip()
        self.logger.debug(f"Generated 'no answer' response: {human_answer}")
        return human_answer


finalize_answer_tmplt = """\
Сформулируй ответ на запрос пользователя основе информации, полученной в результате вызова результатов инструментов.
Если информации недостаточно для точного ответа, ясно опиши ограничения и предложи 1–2 уточняющих вопроса.
Важно! Не упоминай инструменты в явном виде, ссылайся только на данные.
"""

generate_no_answer_tmplt = """\
Ты - помощник, который преобразует технические ответы в понятный человеку текст.
Мы не смогли найти ответ на вопрос пользователя в базе знаний.
Сформулируй ответ, сообщающий кратко и информативно, что ответа не найдено.
Предложи пару вариантов уточняющих вопросов на основе информации в контексте. Предложи в casual стиле.
"""

generate_answer_with_tools_tmplt = """\
Ты — помощник по работе с графовыми базами данных, в которых используется язык запросов Cypher

Цель: постараться найти ответ на вопрос пользователя используя инструменты для работы с БД на основе текстового описания графовой базы данных.

На вход ты получаешь graph_composition: – описание графа и примеры запросов по нему, и user_query – пользовательский запрос.

**Что нужно сделать:**
1. Сгенерировать `Cypher`-запросы, используя узлы, атрибуты и связи перечисленные в **graph_composition**.
2. Руководствуйся данными в **graph_composition** примерами запросов, чтобы составить итоговый запрос.
3. Используй инструменты {tools} для выполнения запросов и поиска

Если нужно, используй несколько `MATCH`-блоков, например:
    MATCH (o:offer)-[:OFFER_belongs_to_CATEGORY]->(c:category)
    MATCH (o)-[:OFFER_made_of_MATERIAL]->(m:material)
    WHERE c.category_name = "Встраиваемый светильник" AND m.material_name IN ["Стекло", "Металл и Стекло", "Алюминий и стекло"]
    RETURN o

Теперь проанализируй следующую структуру графа, и постарайся найти ответ на вопрос используя инструменты {tools}. (Лучше использовать несколько инструментов)

**graph_composition**
{graph_description}
"""


def make_cypher_query_with_tools_dialog(
    graph_description: str,
    prompt_templates: dict[str, str],
    messages: Iterable,
    tool_names: list[str],
) -> list[ChatCompletionMessageParam]:
    prompt_template = prompt_templates.get("generate_answer_with_tools_tmplt", generate_answer_with_tools_tmplt)
    prompt = prompt_template.format(graph_description=graph_description, tools=", ".join(tool_names))
    return [
        {
            "role": "system",
            "content": prompt,
        },
        *messages,
    ]



import asyncio
from functools import cache

import sqlalchemy as sa
import sqlalchemy.ext.asyncio as sa_aio
from pydantic_settings import BaseSettings, SettingsConfigDict


class DbSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="JIMS_", env_file=".env", env_file_encoding="utf-8", extra="ignore")

    db_conn_uri: str = "postgresql://postgres:postgres@localhost:5432"


db_settings = DbSettings()  # type: ignore


# This is needed because each async loop needs its own engine
@cache
def _create_async_engine(loop):
    return sa_aio.create_async_engine(
        db_settings.db_conn_uri.replace("postgresql://", "postgresql+asyncpg://").replace(
            "sqlite://", "sqlite+aiosqlite://"
        )
    )


def get_async_db_engine() -> sa_aio.AsyncEngine:
    return _create_async_engine(asyncio.get_event_loop())


def get_db_engine() -> sa.Engine:
    return sa.create_engine(db_settings.db_conn_uri)


def get_sessionmaker() -> sa_aio.async_sessionmaker[sa_aio.AsyncSession]:
    return sa_aio.async_sessionmaker(
        bind=get_async_db_engine(),
        expire_on_commit=False,
        future=True,
    )



def cast_dtype(value, value_name: str, dtype: str | None):
    """
    умнейшая конвертация строковых переменных по заданному в дата-модели типу данных
    """

    try:
        if dtype == "int" or dtype == "float":
            if isinstance(value, str):
                if len(value.strip(",").split(",")) == 2:
                    value = value.replace(",", ".")
                elif len(value.strip(",").split(",")) > 2:  # assume "," is 000 delimeter
                    value = value.replace(",", "")
                elif len(value.strip(".").split(".")) > 2:  # assume "." is 000 delimeter
                    value = value.replace(".", "")

            value = float(value)
            if dtype == "int":
                value = int(value)

        elif dtype == "bool":
            return str(value).strip().lower() in ["1", "true", "да", "есть"]
        elif dtype == "str":
            return str(value)

        else:
            return value  # Fallback

    except (ValueError, TypeError):
        print(f"cast_dtype error for value {value_name} {value} (expected dtype {dtype})")

    return value



import pytest
import pytest_asyncio
import sqlalchemy as sa
import sqlalchemy.ext.asyncio as sa_aio
from sqlalchemy.pool import StaticPool

from vedana_core.data_model import DataModel

CREATE_REQUIRED_TABLES_SQL = [
    """
    CREATE TABLE dm_anchors (
        noun TEXT PRIMARY KEY,
        description TEXT,
        id_example TEXT,
        query TEXT
    )
    """,
    """
    CREATE TABLE dm_links (
        anchor1 TEXT NOT NULL,
        anchor2 TEXT NOT NULL,
        sentence TEXT NOT NULL,
        description TEXT,
        query TEXT,
        anchor1_link_column_name TEXT,
        anchor2_link_column_name TEXT,
        has_direction BOOLEAN,
        PRIMARY KEY (anchor1, anchor2, sentence)
    )
    """,
    """
    CREATE TABLE dm_anchor_attributes (
        anchor TEXT NOT NULL,
        attribute_name TEXT NOT NULL,
        description TEXT,
        data_example TEXT,
        embeddable BOOLEAN,
        query TEXT,
        dtype TEXT,
        embed_threshold REAL,
        PRIMARY KEY (anchor, attribute_name)
    )
    """,
    """
    CREATE TABLE dm_link_attributes (
        link TEXT NOT NULL,
        attribute_name TEXT NOT NULL,
        description TEXT,
        data_example TEXT,
        embeddable BOOLEAN,
        query TEXT,
        dtype TEXT,
        embed_threshold REAL,
        PRIMARY KEY (link, attribute_name)
    )
    """,
]

CREATE_OPTIONAL_TABLES_SQL = [
    """
    CREATE TABLE dm_queries (
        query_name TEXT PRIMARY KEY,
        query_example TEXT
    )
    """,
    """
    CREATE TABLE dm_prompts (
        name TEXT PRIMARY KEY,
        text TEXT
    )
    """,
    """
    CREATE TABLE dm_conversation_lifecycle (
        event TEXT PRIMARY KEY,
        text TEXT
    )
    """,
]

INSERT_TEST_DATA_SQL = [
    """
    INSERT INTO dm_anchors (noun, description, id_example, query) VALUES
    ('document', 'document', 'document:doc42', ''),
    ('document_chunk', 'document chunk', 'document_chunk_id:doc42_001',
     'MATCH (d:document_chunk) WHERE ID(d)="document_chunk_id:doc42_001" RETURN d'),
    ('consultation', 'main anchor', '',
     'MATCH (c:consultation) WHERE c.consultation_text CONTAINS "keyword" RETURN c, c.consultation_text'),
    ('faq', 'frequent questions and answers', 'answer:001', '')
    """,
    """
    INSERT INTO dm_links (anchor1, anchor2, sentence, description, query,
                        anchor1_link_column_name, anchor2_link_column_name, has_direction) VALUES
    ('document', 'document_chunk', 'DOCUMENT_has_DOCUMENT_CHUNK', '',
     'MATCH (d:document)-[rel:DOCUMENT_has_DOCUMENT_CHUNK]->(c:document_chunk) RETURN d, c',
     'link_document_has_document_chunk', 'link_document_has_document_chunk', 0),
    ('document_chunk', 'faq', 'DOCUMENT_CHUNK_has_FAQ', '',
     'MATCH (d:document_chunk)-[rel:DOCUMENT_CHUNK_has_FAQ]->(f:faq) RETURN d, f',
     'link_document_chunk_has_faq', 'link_document_chunk_has_faq', 0)
    """,
    """
    INSERT INTO dm_anchor_attributes (anchor, attribute_name, description, data_example,
                                     embeddable, query, dtype, embed_threshold) VALUES
    ('consultation', 'consultation_text', 'text', '', 1,
     'vts_fn(label=''consultation'', property=''consultation_text'', text=''<user question>'')',
     'str', 0.05),
    ('faq', 'faq_question_text', 'extract topic and search using faq_question_text', '', 1,
     'vts_fn(label=''faq'', property=''faq_question_text'', text=''<entity>'')', 'str', 0.01),
    ('faq', 'faq_answer_text', '', '', 1, '', 'str', 0.2)
    """,
    """
    INSERT INTO dm_link_attributes (link, attribute_name, description, data_example,
                                   embeddable, query, dtype, embed_threshold) VALUES
    ('DOCUMENT_has_DOCUMENT_CHUNK', 'chunk_order', 'order of chunk in document', '1', 0,
     '', 'int', 0.0),
    ('DOCUMENT_CHUNK_has_FAQ', 'relevance_score', 'relevance score', '0.85', 1,
     '', 'float', 0.1)
    """,
    """
    INSERT INTO dm_queries (query_name, query_example) VALUES
    ('ANY QUESTION', 'extract topics, search FAQs, then search consultations')
    """,
    """
    INSERT INTO dm_prompts (name, text) VALUES
    ('dm_descr_template',
     '## Nodes:\\n{anchors}\\n\\n## Node attributes:\\n{anchor_attrs}\\n\\n## Links:\\n{links}\\n\\n## Link attributes:\\n{link_attrs}\\n\\n## Frequent questions:\\n{queries}'),
    ('generate_no_answer_tmplt', 'Tell the user we could not find the answer.')
    """,
    """
    INSERT INTO dm_conversation_lifecycle (event, text) VALUES
    ('/start', 'Hello!')
    """,
]


@pytest_asyncio.fixture
async def test_db_engine():
    engine = sa_aio.create_async_engine(
        "sqlite+aiosqlite:///:memory:",
        connect_args={"check_same_thread": False},
        poolclass=StaticPool,
    )

    async with engine.begin() as conn:
        for statement in CREATE_REQUIRED_TABLES_SQL + CREATE_OPTIONAL_TABLES_SQL:
            await conn.execute(sa.text(statement))

    yield engine

    await engine.dispose()


@pytest.fixture
def test_sessionmaker(
    test_db_engine,
) -> sa_aio.async_sessionmaker[sa_aio.AsyncSession]:
    return sa_aio.async_sessionmaker(bind=test_db_engine, expire_on_commit=False)


@pytest_asyncio.fixture
async def test_data(test_db_engine):
    """Insert test data into the test database."""
    async with test_db_engine.begin() as conn:
        for statement in INSERT_TEST_DATA_SQL:
            await conn.execute(sa.text(statement))

    yield


@pytest.mark.asyncio
async def test_dm_usage(test_sessionmaker, test_data):
    """Test DataModel reading from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    vis = set(await dm.vector_indices())
    assert ("anchor", "consultation", "consultation_text", 0.05) in vis
    assert ("anchor", "faq", "faq_question_text", 0.01) in vis
    assert ("anchor", "faq", "faq_answer_text", 0.2) in vis
    assert ("edge", "DOCUMENT_CHUNK_has_FAQ", "relevance_score", 0.1) in vis

    doc_links = {link.sentence for link in await dm.anchor_links("document")}
    assert "DOCUMENT_has_DOCUMENT_CHUNK" in doc_links

    tmpls = await dm.prompt_templates()
    assert "dm_descr_template" in tmpls
    assert "generate_no_answer_tmplt" in tmpls

    lifecycle = await dm.conversation_lifecycle_events()
    assert lifecycle.get("/start") == "Hello!"

    text_descr = await dm.to_text_descr()
    assert "consultation_text" in text_descr
    assert "DOCUMENT_has_DOCUMENT_CHUNK" in text_descr


@pytest.mark.asyncio
async def test_dm_anchors(test_sessionmaker, test_data):
    """Test reading anchors from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    anchors = await dm.get_anchors()
    anchor_nouns = {anchor.noun for anchor in anchors}

    assert len(anchors) == 4
    assert "document" in anchor_nouns
    assert "document_chunk" in anchor_nouns
    assert "consultation" in anchor_nouns
    assert "faq" in anchor_nouns

    consultation_anchor = next(a for a in anchors if a.noun == "consultation")
    assert consultation_anchor.description == "main anchor"

    consultation_attrs = [attr for attr in consultation_anchor.attributes if attr.name == "consultation_text"]
    assert len(consultation_attrs) == 1
    assert consultation_attrs[0].embed_threshold == 0.05

    faq_anchor = next(a for a in anchors if a.noun == "faq")
    assert len(faq_anchor.attributes) == 2
    faq_attr_names = {attr.name for attr in faq_anchor.attributes}
    assert "faq_question_text" in faq_attr_names
    assert "faq_answer_text" in faq_attr_names

    document_anchor = next(a for a in anchors if a.noun == "document")
    assert len(document_anchor.attributes) == 0


@pytest.mark.asyncio
async def test_dm_links(test_sessionmaker, test_data):
    """Test reading links from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    anchors = await dm.get_anchors()
    links = await dm.get_links(anchors_dict={anchor.noun: anchor for anchor in anchors})
    link_sentences = {link.sentence for link in links}

    assert len(links) == 2
    assert "DOCUMENT_has_DOCUMENT_CHUNK" in link_sentences
    assert "DOCUMENT_CHUNK_has_FAQ" in link_sentences

    doc_link = next(link for link in links if link.sentence == "DOCUMENT_has_DOCUMENT_CHUNK")
    assert doc_link.anchor_from.noun == "document"
    assert doc_link.anchor_to.noun == "document_chunk"
    assert doc_link.has_direction is False

    assert len(doc_link.attributes) == 1
    assert doc_link.attributes[0].name == "chunk_order"
    assert doc_link.attributes[0].dtype == "int"

    faq_link = next(link for link in links if link.sentence == "DOCUMENT_CHUNK_has_FAQ")
    assert len(faq_link.attributes) == 1
    assert faq_link.attributes[0].name == "relevance_score"
    assert faq_link.attributes[0].embeddable is True
    assert faq_link.attributes[0].embed_threshold == 0.1


@pytest.mark.asyncio
async def test_dm_attributes(test_sessionmaker, test_data):
    """Test reading attributes from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    anchors = await dm.get_anchors()
    links = await dm.get_links(anchors_dict={anchor.noun: anchor for anchor in anchors})

    anchor_attrs = []
    for anchor in anchors:
        anchor_attrs.extend(anchor.attributes)

    anchor_attr_names = {attr.name for attr in anchor_attrs}

    assert len(anchor_attrs) == 3
    assert "consultation_text" in anchor_attr_names
    assert "faq_question_text" in anchor_attr_names
    assert "faq_answer_text" in anchor_attr_names

    consultation_attr = next(a for a in anchor_attrs if a.name == "consultation_text")
    assert consultation_attr.embeddable is True
    assert consultation_attr.embed_threshold == 0.05
    assert consultation_attr.dtype == "str"

    consultation_anchor = next(a for a in anchors if a.noun == "consultation")
    assert any(attr.name == "consultation_text" for attr in consultation_anchor.attributes)

    faq_anchor = next(a for a in anchors if a.noun == "faq")
    assert any(attr.name == "faq_question_text" for attr in faq_anchor.attributes)
    assert any(attr.name == "faq_answer_text" for attr in faq_anchor.attributes)

    link_attrs = []
    for link in links:
        link_attrs.extend(link.attributes)

    link_attr_names = {attr.name for attr in link_attrs}

    assert len(link_attrs) == 2
    assert "chunk_order" in link_attr_names
    assert "relevance_score" in link_attr_names

    doc_link = next(link for link in links if link.sentence == "DOCUMENT_has_DOCUMENT_CHUNK")
    assert any(attr.name == "chunk_order" for attr in doc_link.attributes)

    faq_link = next(link for link in links if link.sentence == "DOCUMENT_CHUNK_has_FAQ")
    assert any(attr.name == "relevance_score" for attr in faq_link.attributes)


@pytest.mark.asyncio
async def test_dm_queries(test_sessionmaker, test_data):
    """Test reading queries from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    queries = await dm.get_queries()
    assert len(queries) == 1
    assert queries[0].name == "ANY QUESTION"
    assert "extract topics" in queries[0].example


@pytest.mark.asyncio
async def test_dm_prompts(test_sessionmaker, test_data):
    """Test reading prompts from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    prompts = await dm.get_prompts()
    prompt_names = {p.name for p in prompts}

    assert len(prompts) == 2
    assert "dm_descr_template" in prompt_names
    assert "generate_no_answer_tmplt" in prompt_names

    templates = await dm.prompt_templates()
    assert "dm_descr_template" in templates
    assert "## Nodes:" in templates["dm_descr_template"]


@pytest.mark.asyncio
async def test_dm_conversation_lifecycle(test_sessionmaker, test_data):
    """Test reading conversation lifecycle events from database."""
    dm = DataModel.create(sessionmaker=test_sessionmaker)

    lifecycle = await dm.get_conversation_lifecycle_events()
    assert len(lifecycle) == 1
    assert lifecycle[0].event == "/start"
    assert lifecycle[0].text == "Hello!"

    events = await dm.conversation_lifecycle_events()
    assert events.get("/start") == "Hello!"


@pytest.mark.asyncio
async def test_dm_empty_optional_tables():
    """Test that optional tables return empty lists if they don't exist or are empty."""
    engine = sa_aio.create_async_engine(
        "sqlite+aiosqlite:///:memory:",
        connect_args={"check_same_thread": False},
        poolclass=StaticPool,
    )
    try:
        async with engine.begin() as conn:
            for statement in CREATE_REQUIRED_TABLES_SQL:
                await conn.execute(sa.text(statement))

        sessionmaker = sa_aio.async_sessionmaker(bind=engine, expire_on_commit=False)
        dm = DataModel.create(sessionmaker=sessionmaker)

        assert await dm.get_queries() == []
        assert await dm.get_prompts() == []
        assert await dm.get_conversation_lifecycle_events() == []
    finally:
        await engine.dispose()



import logging
from dataclasses import dataclass

import sqlalchemy.ext.asyncio as sa_aio
from sqlalchemy import select
from vedana_etl.catalog import (
    dm_anchor_attributes,
    dm_anchors,
    dm_conversation_lifecycle,
    dm_link_attributes,
    dm_links,
    dm_prompts,
    dm_queries,
)

logger = logging.getLogger(__name__)


@dataclass
class Attribute:
    name: str
    description: str
    example: str
    dtype: str
    query: str
    embeddable: bool = False
    embed_threshold: float = 0


@dataclass
class Anchor:
    noun: str
    description: str
    id_example: str
    query: str
    attributes: list[Attribute]

    def __str__(self) -> str:
        return self.noun


@dataclass
class Link:
    anchor_from: Anchor
    anchor_to: Anchor
    sentence: str
    description: str
    query: str
    attributes: list[Attribute]
    has_direction: bool = False
    anchor_from_link_attr_name: str = ""
    anchor_to_link_attr_name: str = ""


@dataclass
class Query:
    name: str
    example: str


@dataclass
class ConversationLifecycleEvent:
    event: str
    text: str


@dataclass
class Prompt:
    name: str
    text: str


class DataModel:
    """
    DataModel, read from SQL tables at runtime
    """

    def __init__(self, sessionmaker: sa_aio.async_sessionmaker[sa_aio.AsyncSession]) -> None:
        self.sessionmaker = sessionmaker

    @classmethod
    def create(cls, sessionmaker) -> "DataModel":
        return cls(sessionmaker=sessionmaker)

    async def get_anchors(self) -> list[Anchor]:
        """Read anchors from dm_anchors table."""
        # .data_table is TableStoreDB's attribute
        anchors_table = dm_anchors.store.data_table  # type: ignore[attr-defined]
        anchors_attr_table = dm_anchor_attributes.store.data_table  # type: ignore[attr-defined]

        async with self.sessionmaker() as session:
            join_query = select(
                anchors_table.c.noun,
                anchors_table.c.description.label("anchor_description"),
                anchors_table.c.id_example,
                anchors_table.c.query.label("anchor_query"),
                anchors_attr_table.c.attribute_name,
                anchors_attr_table.c.description.label("attr_description"),
                anchors_attr_table.c.data_example,
                anchors_attr_table.c.embeddable,
                anchors_attr_table.c.query.label("attr_query"),
                anchors_attr_table.c.dtype,
                anchors_attr_table.c.embed_threshold,
            ).select_from(
                anchors_table.join(  # left join
                    anchors_attr_table,
                    anchors_table.c.noun == anchors_attr_table.c.anchor,
                    isouter=True,
                )
            )
            result = (await session.execute(join_query)).fetchall()

            anchors = {}
            for row in result:
                noun = row.noun
                if noun not in anchors:
                    anchors[noun] = Anchor(
                        noun=noun,
                        description=row.anchor_description,
                        id_example=row.id_example,
                        query=row.anchor_query,
                        attributes=[],
                    )

                # Add attribute if it exists (attribute_name will be None for anchors without attributes)
                if row.attribute_name is not None:
                    anchors[noun].attributes.append(
                        Attribute(
                            name=row.attribute_name,
                            description=row.attr_description if row.attr_description else "",
                            example=row.data_example if row.data_example else "",
                            embeddable=row.embeddable if row.embeddable is not None else False,
                            query=row.attr_query if row.attr_query else "",
                            dtype=row.dtype if row.dtype else "",
                            embed_threshold=row.embed_threshold if row.embed_threshold is not None else 1.0,
                        )
                    )

            return list(anchors.values())

    async def get_links(self, anchors_dict: dict[str, Anchor] | None = None) -> list[Link]:
        """Read links from dm_links table."""
        links_table = dm_links.store.data_table  # type: ignore[attr-defined]
        links_attr_table = dm_link_attributes.store.data_table  # type: ignore[attr-defined]

        if anchors_dict is None:
            anchors = await self.get_anchors()
            anchors_dict = {anchor.noun: anchor for anchor in anchors}

        async with self.sessionmaker() as session:
            join_query = select(
                links_table.c.anchor1,
                links_table.c.anchor2,
                links_table.c.sentence,
                links_table.c.description.label("link_description"),
                links_table.c.query.label("link_query"),
                links_table.c.anchor1_link_column_name,
                links_table.c.anchor2_link_column_name,
                links_table.c.has_direction,
                links_attr_table.c.attribute_name,
                links_attr_table.c.description.label("attr_description"),
                links_attr_table.c.data_example,
                links_attr_table.c.embeddable,
                links_attr_table.c.query.label("attr_query"),
                links_attr_table.c.dtype,
                links_attr_table.c.embed_threshold,
            ).select_from(
                links_table.join(  # left join
                    links_attr_table,
                    links_table.c.sentence == links_attr_table.c.link,
                    isouter=True,
                )
            )

            result = (await session.execute(join_query)).fetchall()

            links = {}
            for row in result:
                sentence = row.sentence
                if sentence not in links:
                    anchor_from = anchors_dict.get(row.anchor1)
                    anchor_to = anchors_dict.get(row.anchor2)
                    if anchor_from is None or anchor_to is None:
                        logger.error(f'Link {sentence} has invalid connection "{row.anchor1} - {row.anchor2}"')
                        continue

                    links[sentence] = Link(
                        anchor_from=anchor_from,
                        anchor_to=anchor_to,
                        anchor_from_link_attr_name=row.anchor1_link_column_name,
                        anchor_to_link_attr_name=row.anchor2_link_column_name,
                        sentence=sentence,
                        description=row.link_description,
                        query=row.link_query,
                        has_direction=bool(row.has_direction) if row.has_direction is not None else False,
                        attributes=[],
                    )

                # Add attribute if it exists (attribute_name will be None for anchors without attributes)
                if row.attribute_name is not None:
                    links[sentence].attributes.append(
                        Attribute(
                            name=row.attribute_name,
                            description=row.attr_description if row.attr_description else "",
                            example=row.data_example if row.data_example else "",
                            embeddable=row.embeddable if row.embeddable is not None else False,
                            query=row.attr_query if row.attr_query else "",
                            dtype=row.dtype if row.dtype else "",
                            embed_threshold=row.embed_threshold if row.embed_threshold is not None else 1.0,
                        )
                    )

            return list(links.values())

    async def get_queries(self) -> list[Query]:
        try:
            queries_table = dm_queries.store.data_table  # type: ignore[attr-defined]
            async with self.sessionmaker() as session:
                result = (await session.execute(select(queries_table))).fetchall()
                return [Query(name=row.query_name, example=row.query_example) for row in result]
        except Exception:
            return []

    async def get_conversation_lifecycle_events(self) -> list[ConversationLifecycleEvent]:
        try:
            lifecycle_table = dm_conversation_lifecycle.store.data_table  # type: ignore[attr-defined]
            async with self.sessionmaker() as session:
                result = (await session.execute(select(lifecycle_table))).fetchall()
                return [ConversationLifecycleEvent(event=row.event, text=row.text) for row in result]
        except Exception:
            return []

    async def conversation_lifecycle_events(self) -> dict[str, str]:
        cl = await self.get_conversation_lifecycle_events()
        return {c.event: c.text for c in cl}

    async def get_prompts(self) -> list[Prompt]:
        try:
            prompts_table = dm_prompts.store.data_table  # type: ignore[attr-defined]
            async with self.sessionmaker() as session:
                result = (await session.execute(select(prompts_table))).fetchall()
                return [Prompt(name=row.name, text=row.text) for row in result]
        except Exception:
            return []

    async def prompt_templates(self) -> dict[str, str]:
        prompts = await self.get_prompts()
        return {p.name: p.text for p in prompts}

    async def vector_indices(self) -> list[tuple[str, str, str, float]]:
        """
        returns list
        ("anchor", anchor.noun, anchor.attribute, anchor.th) +
        ("edge", link.sentence, link.attribute, link.th)
        for all embeddable attributes
        """
        anchors = await self.get_anchors()
        links = await self.get_links(anchors_dict={a.noun: a for a in anchors})

        a_i = [
            ("anchor", anchor.noun, attr.name, attr.embed_threshold)
            for anchor in anchors
            for attr in anchor.attributes
            if attr.embeddable
        ]
        l_i = [
            ("edge", link.sentence, attr.name, attr.embed_threshold)
            for link in links
            for attr in link.attributes
            if attr.embeddable
        ]
        return a_i + l_i

    async def anchor_links(self, anchor_noun: str) -> list[Link]:
        """all links that connect to/from this anchor"""
        links = await self.get_links()
        return [
            link
            for link in links
            if (link.anchor_from.noun == anchor_noun and link.anchor_from_link_attr_name)
            or (link.anchor_to.noun == anchor_noun and link.anchor_to_link_attr_name)
        ]

    async def to_text_descr(
        self,
        anchor_nouns: list[str] | None = None,
        link_sentences: list[str] | None = None,
        anchor_attribute_names: list[str] | None = None,
        link_attribute_names: list[str] | None = None,
        query_names: list[str] | None = None,
    ) -> str:
        """Create a text description of the data model, optionally filtered.

        Args:
            anchor_nouns: List of anchor nouns to include. If None, includes all.
            link_sentences: List of link sentences to include. If None, includes all.
            anchor_attribute_names: List of anchor attribute names to include. If None, includes all.
            link_attribute_names: List of link attribute names to include. If None, includes all.
            query_names: List of query names to include. If None, includes all.

        Returns:
            A formatted string description of the data model.
        """
        anchors = await self.get_anchors()
        links = await self.get_links(anchors_dict={a.noun: a for a in anchors})
        queries = await self.get_queries()
        dm_templates = await self.prompt_templates()

        # Convert to sets for efficient lookup, None means include all
        anchor_set = set(anchor_nouns) if anchor_nouns is not None else None
        link_set = set(link_sentences) if link_sentences is not None else None
        anchor_attr_set = set(anchor_attribute_names) if anchor_attribute_names is not None else None
        link_attr_set = set(link_attribute_names) if link_attribute_names is not None else None
        query_set = set(query_names) if query_names is not None else None

        # Filter anchors
        filtered_anchors = [anchor for anchor in anchors if anchor_set is None or anchor.noun in anchor_set]

        # Create a map for quick anchor lookup (for link filtering)
        anchors_map = {a.noun: a for a in filtered_anchors}

        # Filter links (only include if both anchors are in filtered set)
        filtered_links = [
            link
            for link in links
            if (link_set is None or link.sentence in link_set)
            and link.anchor_from.noun in anchors_map
            and link.anchor_to.noun in anchors_map
        ]

        anchor_descr = "\n".join(
            dm_templates.get("dm_anchor_descr_template", dm_anchor_descr_template).format(anchor=anchor)
            for anchor in filtered_anchors
        )

        anchor_attrs_descr = "\n".join(
            dm_templates.get("dm_attr_descr_template", dm_attr_descr_template).format(anchor=anchor, attr=attr)
            for anchor in filtered_anchors
            for attr in anchor.attributes
            if anchor_attr_set is None or attr.name in anchor_attr_set
        )

        link_descr = "\n".join(
            dm_templates.get("dm_link_descr_template", dm_link_descr_template).format(link=link)
            for link in filtered_links
        )

        link_attrs_descr = "\n".join(
            dm_templates.get("dm_link_attr_descr_template", dm_link_attr_descr_template).format(link=link, attr=attr)
            for link in filtered_links
            for attr in link.attributes
            if link_attr_set is None or attr.name in link_attr_set
        )

        filtered_queries = [q for q in queries if query_set is None or q.name in query_set]
        queries_descr = "\n".join(
            dm_templates.get("dm_query_descr_template", dm_query_descr_template).format(query=query)
            for query in filtered_queries
        )

        dm_template = dm_templates.get("dm_descr_template", dm_descr_template)

        return dm_template.format(
            anchors=anchor_descr,
            anchor_attrs=anchor_attrs_descr,
            links=link_descr,
            link_attrs=link_attrs_descr,
            queries=queries_descr,
        )

    async def to_compact_json(self) -> dict:
        anchors = await self.get_anchors()
        links = await self.get_links(anchors_dict={a.noun: a for a in anchors})
        queries = await self.get_queries()

        descr = {
            "anchors": [
                {
                    "name": a.noun,
                    "description": a.description,
                    "example": a.id_example,
                    "attributes": [
                        {
                            "attr_name": aa.name,
                            "attr_description": aa.description,
                        }
                        for aa in a.attributes
                    ],
                }
                for a in anchors
            ],
            "links": [
                {
                    "from": li.anchor_from,
                    "to": li.anchor_to,
                    "sentence": li.sentence,
                    "description": li.description,
                    "attributes": [
                        {
                            "attr_name": la.name,
                            "attr_description": la.description,
                        }
                        for la in li.attributes
                    ],
                }
                for li in links
            ],
            "queries": {i: q.name for i, q in enumerate(queries, start=1)},
        }
        return descr


# default templates
dm_descr_template = """\
## Узлы:
{anchors}

## Атрибуты узлов:
{anchor_attrs}

## Связи между узлами:
{links}

## Атрибуты связей:
{link_attrs}

## Типичные вопросы:
{queries}
"""

dm_anchor_descr_template = (
    "- {anchor.noun}: {anchor.description}; пример ID: {anchor.id_example}; запрос для получения: {anchor.query}"
)
dm_attr_descr_template = (
    "- {anchor.noun}.{attr.name}: {attr.description}; пример: {attr.example}; запрос для получения: {attr.query}"
)
dm_link_descr_template = "- {link.sentence}: {link.description}; пример запроса: {link.query}"
dm_link_attr_descr_template = (
    "- {link.sentence}.{attr.name}: {attr.description}; пример: {attr.example}; запрос для получения: {attr.query}"
)
dm_query_descr_template = "- {query.name}\n{query.example}"

# Compact templates (without cypher queries)
dm_compact_descr_template = """\
## Узлы:
{anchors}

## Атрибуты узлов:
{anchor_attrs}

## Связи между узлами:
{links}

## Атрибуты связей:
{link_attrs}

## Сценарии вопросов:
{queries}
"""

dm_compact_anchor_descr_template = "- {anchor.noun}: {anchor.description}"
dm_compact_attr_descr_template = "- {anchor.noun}.{attr.name}: {attr.description}"
dm_compact_link_descr_template = "- {link.sentence}: {link.description}"
dm_compact_link_attr_descr_template = "- {link.sentence}.{attr.name}: {attr.description}"
dm_compact_query_descr_template = "- {query.name}"



import enum
import json
import logging
import re
from dataclasses import dataclass
from itertools import islice
from typing import Any, Mapping, Type

import neo4j
import neo4j.graph
from jims_core.thread.thread_context import ThreadContext
from pydantic import BaseModel, Field, create_model

from vedana_core.graph import Graph, Record
from vedana_core.vts import VectorStore
from vedana_core.llm import LLM, Tool
from vedana_core.settings import settings

QueryResult = list[Record] | Exception


# TODO replace with VTSArgs and CypherArgs
class CypherQuery(str):
    ...


@dataclass
class VTSQuery:
    label: str
    param: str
    query: str

    def __str__(self) -> str:
        return f'vector_search("{self.label}","{self.param}","{self.query}")'


DBQuery = CypherQuery | VTSQuery | str


@dataclass
class RagResults:
    fts_res: list[Record] | None = None
    vts_res: list[Record] | None = None
    db_query_res: list[tuple[DBQuery, QueryResult]] | None = None


class VTSArgs(BaseModel):
    label: str = Field(description="node label")
    property: str = Field(description="node property to search in")
    text: str = Field(description="text to search similar")


class CypherArgs(BaseModel):
    query: str = Field(description="Cypher query")


class GetHistoryArgs(BaseModel):
    max_history: int = Field(20000, description="Maximum text length to retrieve from history. Cuts off on messages")


VTS_TOOL_NAME = "vector_text_search"
CYPHER_TOOL_NAME = "cypher"


class RagAgent:
    _vts_args: type[VTSArgs]

    def __init__(
        self,
        graph: Graph,
        vts: VectorStore,
        data_model_description: str,
        data_model_vts_indices,
        llm: LLM,
        ctx: ThreadContext,
        logger: logging.Logger | None = None,
    ) -> None:
        self.graph = graph
        self.vts = vts
        self.llm = llm
        self.logger = logger or logging.getLogger(__name__)
        self._vts_meta_args: dict[str, dict[str, tuple[str, float]]] = {}  # stuff not passed through toolcall
        self._vts_args = self._build_vts_arg_model(data_model_vts_indices)
        self.data_model_description = data_model_description
        self.ctx = ctx

    def _build_vts_arg_model(self, vts_indices) -> Type[VTSArgs]:
        """Create a Pydantic model with Enum-constrained fields for the VTS tool."""

        if not vts_indices:
            return VTSArgs

        # fill in lookup for resolving idx type and getting custom threshold
        for idx_type, i_name, i_attr, i_th in vts_indices:
            if not self._vts_meta_args.get(i_name):
                self._vts_meta_args[i_name] = {}
            self._vts_meta_args[i_name][i_attr] = (idx_type, i_th)

        # Label Enum – keys of `vts_indices`
        LabelEnum = enum.Enum("LabelEnum", {name: name for (_type, name, _attr, _th) in vts_indices})  # type: ignore

        # Property Enum – unique values of `vts_indices`
        unique_props = set(attr for (_type, _name, attr, _th) in vts_indices)
        prop_member_mapping: dict[str, str] = {}

        used_names: set[str] = set()
        for idx, prop in enumerate(sorted(unique_props)):
            sanitized = re.sub(r"\W|^(?=\d)", "_", prop)
            if sanitized in used_names:
                sanitized = f"{sanitized}_{idx}"
            used_names.add(sanitized)
            prop_member_mapping[sanitized] = prop

        PropertyEnum = enum.Enum("PropertyEnum", prop_member_mapping)  # type: ignore

        VTSArgsEnum = create_model(
            "VTSArgsEnum",
            label=(LabelEnum, Field(description="node or edge label")),
            property=(PropertyEnum, Field(description="property to search in")),
            text=(str, Field(description="text for semantic search")),
            __base__=VTSArgs,
        )

        return VTSArgsEnum

    async def search_vector_text(
        self,
        label: str,
        prop_type: str,
        prop_name: str,
        search_value: str,
        threshold: float,
        top_n: int = 5,
    ) -> list[Record]:
        embed = await self.llm.llm.create_embedding(search_value)
        return await self.vts.vector_search(label, prop_type, prop_name, embed, threshold=threshold, top_n=top_n)

    @staticmethod
    def result_to_text(query: str, result: list[Record] | Exception) -> str:
        if isinstance(result, Exception):
            return f"Query: {query}\nResult: 'Error executing query'"
        rows_str = "\n".join(row_to_text(row) for row in result)
        return f"Query: {query}\nRows:\n{rows_str}"

    async def execute_cypher_query(self, query, rows_limit: int = 30) -> QueryResult:
        try:
            return list(islice(await self.graph.execute_ro_cypher_query(query), rows_limit))
        except Exception as e:
            self.logger.exception(e)
            return e

    def rag_results_to_text(self, results: RagResults) -> str:
        all_results = results.db_query_res or []
        if results.vts_res:
            all_results.append(("Vector text search", results.vts_res))
        if results.fts_res:
            all_results.append(("Full text search", results.fts_res))
        return "\n\n".join(self.result_to_text(str(q), r) for q, r in all_results)

    async def text_to_answer_with_vts_and_cypher(
        self, text_query: str, threshold: float, top_n: int = 5
    ) -> tuple[str, list, list[VTSQuery], list[CypherQuery]]:
        vts_queries: list[VTSQuery] = []
        cypher_queries: list[CypherQuery] = []

        async def vts_fn(args: VTSArgs) -> str:
            label = args.label.value if isinstance(args.label, enum.Enum) else args.label
            prop = args.property.value if isinstance(args.property, enum.Enum) else args.property

            # default fallback treats toolcall as node vector search
            prop_type, th = self._vts_meta_args.get(label, {}).get(prop, ("node", threshold))
            self.logger.debug(f"vts_fn(on={prop_type} label={label}, property={prop}, th={th}, n={top_n})")

            vts_queries.append(VTSQuery(label, prop, args.text))
            vts_res = await self.search_vector_text(label, prop_type, prop, args.text, threshold=th, top_n=top_n)
            return self.result_to_text(VTS_TOOL_NAME, vts_res)

        async def cypher_fn(args: CypherArgs) -> str:
            self.logger.debug(f"cypher_fn({args})")
            cypher_queries.append(CypherQuery(args.query))
            res = await self.execute_cypher_query(args.query)
            return self.result_to_text(CYPHER_TOOL_NAME, res)

        vts_tool = Tool(
            VTS_TOOL_NAME,
            "Vector search for similar text in node properties, use for semantic search",
            self._vts_args,
            vts_fn,
        )

        cypher_tool = Tool(
            CYPHER_TOOL_NAME,
            "Execute a Cypher query against the graph database. Use for structured data retrieval and graph traversal.",
            CypherArgs,
            cypher_fn,
        )

        tools: list[Tool] = [vts_tool, cypher_tool]

        all_query_events, answer = await self.llm.generate_cypher_query_with_tools(
            data_descr=self.data_model_description,
            messages=self.ctx.history[-settings.pipeline_history_length :],
            tools=tools,
        )

        if not answer:
            self.logger.warning(f"No answer found for {text_query}. Generating empty answer...")
            answer = await self.llm.generate_no_answer(self.ctx.history[-settings.pipeline_history_length :])

        return answer, all_query_events, vts_queries, cypher_queries


def _remove_embeddings(val: Any):
    if isinstance(val, Mapping):
        return {k: _remove_embeddings(v) for k, v in val.items() if not k.endswith("_embedding")}
    if isinstance(val, list):
        return [_remove_embeddings(v) for v in val]
    return val


def _clear_record_val(val: Any):  # todo check for edges
    params = _remove_embeddings(val)
    if isinstance(val, neo4j.graph.Node) and isinstance(params, dict):
        params["labels"] = list(val.labels)
    return params


def row_to_text(row: Any) -> str:
    if isinstance(row, neo4j.Record):
        row = {k: _clear_record_val(v) for k, v in row.items()}
    try:
        return json.dumps(row, ensure_ascii=False, indent=2)
    except TypeError:
        return str(row)



import logging
import traceback
from dataclasses import asdict
from typing import Any
from pydantic import BaseModel, Field

from jims_core.thread.thread_context import ThreadContext

from vedana_core.data_model import DataModel
from vedana_core.graph import Graph
from vedana_core.vts import VectorStore
from vedana_core.llm import LLM
from vedana_core.rag_agent import RagAgent
from vedana_core.settings import settings


class DataModelSelection(BaseModel):
    reasoning: str = Field(
        default="",
        description="Brief explanation of why these elements were selected for answering the user's question",
    )
    anchor_nouns: list[str] = Field(
        default_factory=list,
        description="List of anchor nouns (node types) needed to answer the question",
    )
    link_sentences: list[str] = Field(
        default_factory=list,
        description="List of link sentences (relationship types) needed to answer the question",
    )
    anchor_attribute_names: list[str] = Field(
        default_factory=list,
        description="List of anchor attribute names needed to answer the question (attributes belonging to nodes)",
    )
    link_attribute_names: list[str] = Field(
        default_factory=list,
        description="List of link attribute names needed to answer the question (attributes belonging to relationships)",
    )
    query_ids: list[str] = Field(
        default_factory=list,
        description="List of query scenario ID's that match the user's question type",
    )


dm_filter_base_system_prompt = """\
Ты — помощник по анализу структуры графовой базы данных.

Твоя задача: проанализировать вопрос пользователя и определить, какие элементы модели данных (узлы, связи, атрибуты, сценарии запросов) необходимы для формирования ответа.

**Правила выбора:**
1. Выбирай только те элементы, которые ДЕЙСТВИТЕЛЬНО нужны для ответа на вопрос
2. Если вопрос касается связи между сущностями — выбери соответствующие узлы И связь между ними
3. Выбирай атрибуты, которые могут содержать искомую информацию или использоваться для фильтрации
   - anchor_attribute_names: атрибуты узлов (находятся в разделе "Атрибуты узлов")
   - link_attribute_names: атрибуты связей (находятся в разделе "Атрибуты связей")
4. Выбирай сценарий запроса, который лучше всего соответствует типу вопроса пользователя
5. Лучше выбрать чуть больше, чем упустить важное — но не выбирай всё подряд

**Формат ответа:**
Верни JSON с выбранными элементами. Используй ТОЧНЫЕ имена из предоставленной модели данных.
"""

dm_filter_user_prompt_template = """\
**Вопрос пользователя:**
{user_query}

**Модель данных:**
{compact_data_model}

Проанализируй вопрос и выбери необходимые элементы модели данных для формирования ответа.
"""


class StartPipeline:
    """
    Response for /start command
    """

    def __init__(self, data_model: DataModel) -> None:
        self.data_model = data_model

    async def __call__(self, ctx: ThreadContext) -> None:
        lifecycle_events = await self.data_model.conversation_lifecycle_events()
        start_response = lifecycle_events.get("/start")

        message = start_response or "Bot online. No response for /start command in LifecycleEvents"
        ctx.send_message(message)


class RagPipeline:
    """RAG Pipeline with data model filtering for optimized query processing.

    This pipeline adds an initial step that filters the data model based on
    the user's query, selecting only relevant anchors, links, attributes,
    and query scenarios. This reduces token usage and improves LLM precision
    for large data models.
    """

    def __init__(
        self,
        graph: Graph,
        vts: VectorStore,
        data_model: DataModel,
        logger,
        threshold: float = 0.8,
        top_n: int = 5,
        model: str | None = None,
        filter_model: str | None = None,
        enable_filtering: bool | None = None,
    ):
        self.graph = graph
        self.vts = vts
        self.data_model = data_model
        self.logger = logger or logging.getLogger(__name__)
        self.threshold = threshold
        self.top_n = top_n
        self.model = model or settings.model
        self.filter_model = filter_model or settings.filter_model  # or self.model
        self.enable_filtering = enable_filtering or settings.enable_dm_filtering

    async def __call__(self, ctx: ThreadContext) -> None:
        """Main pipeline execution - implements JIMS Pipeline protocol."""

        # Get the last user message
        user_query = ctx.get_last_user_message()
        if not user_query:
            ctx.send_message("I didn't receive a question. Please ask me something!")
            return

        try:
            # Update status
            await ctx.update_agent_status("Processing your question...")

            # Process the query using RAG
            answer, agent_query_events, technical_info = await self.process_rag_query(user_query, ctx)

            # Send the answer
            ctx.send_message(answer)

            # Store technical information as an event
            ctx.send_event(
                "rag.query_processed",
                {
                    "query": user_query,
                    "answer": answer,
                    "technical_info": technical_info,
                    "threshold": self.threshold,
                },
            )

        except Exception as e:
            self.logger.exception(f"Error in RAG pipeline: {e}")
            error_msg = "An error occurred while processing the request"  # не передаем ошибку пользователю в диалог
            ctx.send_message(error_msg)

            # Store error event
            ctx.send_event(
                "rag.error",
                {
                    "query": user_query,
                    "error": str(e),
                    "traceback": traceback.format_exc(),
                },
            )

    async def process_rag_query(self, query: str, ctx: ThreadContext) -> tuple[str, list, dict[str, Any]]:
        # 1. Filter data model
        if self.enable_filtering:
            await ctx.update_agent_status("Analyzing query structure...")
            data_model_description, filter_selection = await self.filter_data_model(query, ctx)
        else:
            data_model_description = await self.data_model.to_text_descr()
            filter_selection = DataModelSelection()

        # Read required DataModel properties
        prompt_templates = await self.data_model.prompt_templates()
        data_model_vector_search_indices = await self.data_model.vector_indices()

        # 2. Create LLM and agent with filtered data model; Step 1 LLM costs are counted since same ctx.llm is used
        llm = LLM(ctx.llm, prompt_templates=prompt_templates, logger=self.logger)
        await ctx.update_agent_status("Searching knowledge base...")

        if self.model != llm.llm.model and settings.debug:
            llm.llm.set_model(self.model)

        agent = RagAgent(
            graph=self.graph,
            vts=self.vts,
            data_model_description=data_model_description,
            data_model_vts_indices=data_model_vector_search_indices,
            llm=llm,
            ctx=ctx,
            logger=self.logger,
        )

        (
            answer,
            agent_query_events,
            vts_queries,
            cypher_queries,
        ) = await agent.text_to_answer_with_vts_and_cypher(
            query,
            threshold=self.threshold,
            top_n=self.top_n,
        )

        technical_info: dict[str, Any] = {
            "vts_queries": [str(q) for q in vts_queries],
            "cypher_queries": [str(q) for q in cypher_queries],
            "num_vts_queries": len(vts_queries),
            "num_cypher_queries": len(cypher_queries),
            "model_used": self.model,
            "model_stats": {m: asdict(u) for m, u in ctx.llm.usage.items()},
        }

        # Add filtering info if applicable
        if self.enable_filtering:
            dm_anchors = await self.data_model.get_anchors()
            dm_links = await self.data_model.get_links()
            dm_queries = await self.data_model.get_queries()

            # Count total attributes for original_counts
            total_anchor_attrs = sum(len(anchor.attributes) for anchor in dm_anchors)
            total_link_attrs = sum(len(link.attributes) for link in dm_links)

            dm_filtering = {
                "filter_model": self.filter_model,
                "reasoning": filter_selection.reasoning,
                "selected_anchors": filter_selection.anchor_nouns,
                "selected_links": filter_selection.link_sentences,
                "selected_anchor_attributes": filter_selection.anchor_attribute_names,
                "selected_link_attributes": filter_selection.link_attribute_names,
                "selected_queries": [dm_queries[int(i)].name for i in filter_selection.query_ids],
                "original_counts": {
                    "anchors": len(dm_anchors),
                    "links": len(dm_links),
                    "anchor_attrs": total_anchor_attrs,
                    "link_attrs": total_link_attrs,
                    "queries": len(dm_queries),
                },
                "filtered_counts": {
                    "anchors": len(filter_selection.anchor_nouns),
                    "links": len(filter_selection.link_sentences),
                    "anchor_attrs": len(filter_selection.anchor_attribute_names),
                    "link_attrs": len(filter_selection.link_attribute_names),
                    "queries": len(filter_selection.query_ids),
                },
            }

            ctx.send_event("rag.data_model_filtered", dm_filtering)

        return answer, agent_query_events, technical_info

    async def filter_data_model(
        self,
        query: str,
        ctx: ThreadContext,
    ) -> tuple[str, DataModelSelection]:
        # Get description for filtering
        dm_json = await self.data_model.to_compact_json()
        dm_prompt_templates = await self.data_model.prompt_templates()

        # Build the prompt
        system_prompt = dm_prompt_templates.get("dm_filter_prompt", dm_filter_base_system_prompt)
        user_prompt = dm_prompt_templates.get("dm_filter_user_prompt", dm_filter_user_prompt_template).format(
            user_query=query,
            compact_data_model=dm_json,
        )

        messages: list[dict[str, Any]] = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]

        self.logger.debug(f"Filtering data model for query: {query}")

        try:
            filter_llm = ctx.llm

            base_model = ctx.llm.model
            if self.filter_model:  # if different model specified for filtering - use it
                filter_llm.set_model(self.filter_model)

            # Use structured output to get the selection
            selection = await filter_llm.chat_completion_structured(messages, DataModelSelection)

            if base_model:  # select base model back
                ctx.llm.set_model(base_model)

            if selection is None:
                raise ValueError("LLM returned empty response")

            # parse query id's to query names - LLM often misspells arbitrary query_names
            query_names = [dm_json["queries"].get(int(i)) for i in selection.query_ids]

            self.logger.debug(
                f"Data model filter selection: "
                f"anchors={selection.anchor_nouns}, "
                f"links={selection.link_sentences}, "
                f"anchor_attrs={selection.anchor_attribute_names}, "
                f"link_attrs={selection.link_attribute_names}, "
                f"queries={query_names}",
            )
            self.logger.debug(f"Filter reasoning: {selection.reasoning}")

            # Create filtered data model
            filtered_dm_descr = await self.data_model.to_text_descr(
                anchor_nouns=selection.anchor_nouns,
                link_sentences=selection.link_sentences,
                anchor_attribute_names=selection.anchor_attribute_names,
                link_attribute_names=selection.link_attribute_names,
                query_names=query_names,
            )

            return filtered_dm_descr, selection

        except Exception as e:  # return full data_model
            self.logger.exception(f"Data model filtering failed: {e}. Using full data model.")
            descr = await self.data_model.to_text_descr()
            return descr, DataModelSelection(
                reasoning=f"Filtering failed: {e}. Using full data model.",
                anchor_nouns=[],
                link_sentences=[],
                anchor_attribute_names=[],
                link_attribute_names=[],
                query_ids=[],
            )



from unittest.mock import AsyncMock, MagicMock

import pytest
from jims_core.llms.llm_provider import LLMProvider
from pydantic import BaseModel

from vedana_core.llm import LLM, Tool


def mock_msg(content: str | None = None, tool_calls: list | None = None):
    """Create a mock message with to_dict() method."""
    m = MagicMock(content=content, tool_calls=tool_calls, role="assistant")
    m.to_dict.return_value = {"role": "assistant", "content": content}
    return m


def mock_tool_call(name: str, args: str, id: str = "call_1"):
    """Create a mock tool call."""
    tc = MagicMock(id=id)
    tc.function.name = name
    tc.function.arguments = args
    return tc


@pytest.mark.asyncio
async def test_llm_completion_with_tools() -> None:
    llm_provider = LLMProvider()

    tc = mock_tool_call("hello_world", '{"name": "Alice"}')
    llm_provider.chat_completion_with_tools = AsyncMock(
        side_effect=[(mock_msg(tool_calls=[tc]), [tc]), (mock_msg("Done!"), [])]
    )

    llm = LLM(llm_provider=llm_provider, prompt_templates={})

    class HelloWorldArgs(BaseModel):
        name: str

    res_messages, res_content = await llm.create_completion_with_tools(
        messages=[{"role": "system", "content": "Call hello_world with name='Alice'."}],
        tools=[
            Tool(
                name="hello_world", description="Says hello.", args_cls=HelloWorldArgs, fn=lambda a: f"Hello, {a.name}!"
            )
        ],
    )

    tool_msg = next(m for m in res_messages if m.get("role") == "tool")
    assert tool_msg.get("content") == "Hello, Alice!"


@pytest.mark.asyncio
async def test_llm_completion_no_tool_calls() -> None:
    llm_provider = LLMProvider()
    llm_provider.chat_completion_with_tools = AsyncMock(return_value=(mock_msg("No tools needed."), []))

    llm = LLM(llm_provider=llm_provider, prompt_templates={})

    class DummyArgs(BaseModel):
        value: str

    res_messages, res_content = await llm.create_completion_with_tools(
        messages=[{"role": "user", "content": "Hello"}],
        tools=[Tool(name="dummy", description="Dummy.", args_cls=DummyArgs, fn=lambda a: a.value)],
    )

    assert res_content == "No tools needed."
    assert not any(m.get("role") == "tool" for m in res_messages)



from jims_core.thread.thread_context import ThreadContext
from vedana_core.data_model import DataModel


class StartPipeline:
    """
    Response for /start command
    """

    def __init__(self, data_model: DataModel) -> None:
        self.data_model = data_model

    async def __call__(self, ctx: ThreadContext) -> None:
        lifecycle_events = await self.data_model.conversation_lifecycle_events()
        start_response = lifecycle_events.get("/start")
        message = start_response or "Bot online. No response for /start command in LifecycleEvents"
        ctx.send_message(message)



import abc
import io
import sqlite3
import time
from ast import literal_eval
from dataclasses import dataclass
from pathlib import Path
from typing import Any, NamedTuple, Sequence, Union

import grist_api
import pandas as pd
import requests

from vedana_core.data_model import Attribute, Link
from vedana_core.utils import cast_dtype


@dataclass
class Table:
    columns: list[str]
    rows: Sequence[Union[tuple, NamedTuple]]


@dataclass
class AnchorRecord:
    id: str
    type: str
    data: dict[str, Any]
    dp_id: int | None = None


@dataclass
class LinkRecord:
    id_from: str
    id_to: str
    anchor_from: str
    anchor_to: str
    type: str
    data: dict[str, Any]


class DataProvider:
    def get_anchors(self, type_: str, dm_attrs: list[Attribute], dm_anchor_links: list[Link]) -> list[AnchorRecord]:
        raise NotImplementedError("get_anchors must be implemented in subclass")

    def get_links(self, table_name: str, link: Link) -> list[LinkRecord]:
        raise NotImplementedError("get_links must be implemented in subclass")

    def get_anchor_tables(self) -> list[str]:
        raise NotImplementedError("get_anchor_tables must be implemented in subclass")

    def list_anchor_tables(self) -> list[str]:
        raise NotImplementedError("list_anchor_tables must be implemented in subclass")

    def get_link_tables(self) -> list[str]:
        raise NotImplementedError("get_link_tables must be implemented in subclass")

    def close(self) -> None:
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()


class GristDataProvider(DataProvider):
    anchor_table_prefix = "Anchor_"
    link_table_prefix = "Link_"

    def get_anchor_tables(self) -> list[str]:
        prefix_len = len(self.anchor_table_prefix)
        return [t[prefix_len:] for t in self.list_anchor_tables()]

    @abc.abstractmethod
    def list_anchor_tables(self) -> list[str]:
        ...

    @abc.abstractmethod
    def list_link_tables(self) -> list[str]:
        ...

    @abc.abstractmethod
    def get_table(self, table_name: str) -> pd.DataFrame:
        ...

    def get_anchor_types(self) -> list[str]:
        prefix_len = len(self.anchor_table_prefix)
        return [t[prefix_len:] for t in self.list_anchor_tables()]

    def get_link_tables(self) -> list[str]:
        prefix_len = len(self.link_table_prefix)
        return [t[prefix_len:] for t in self.list_link_tables()]

    def get_anchors(self, type_: str, dm_attrs: list[Attribute], dm_anchor_links: list[Link]) -> list[AnchorRecord]:
        table_name = f"{self.anchor_table_prefix}{type_}"
        table = self.get_table(table_name)
        if "id" not in table.columns:
            table["id"] = [None] * table.shape[0]
        table_records: list[dict] = table.to_dict(orient="records")

        id_key = f"{type_}_id"
        anchor_ids = set()
        anchors: list[AnchorRecord] = []

        def flatten_list_cells(el):
            if isinstance(el, list) and el[0] == "L":  # flatten List type fields
                if len(el) == 2:
                    return el[1]
                return el[1:]
            elif pd.isna(el):
                return None
            return el

        for row_dict in table_records:
            db_id = flatten_list_cells(row_dict.pop("id"))
            id_ = row_dict.pop(id_key, None)
            if id_ is None:
                id_ = f"{type_}:{db_id}"
            if pd.isna(id_):
                # print(f"{type_}:{db_id} has {id_key}=nan, skipping")
                continue

            row_dict = {k: flatten_list_cells(v) for k, v in row_dict.items()}
            row_dict = {k: v for k, v in row_dict.items() if not isinstance(v, (bytes, type(None)))}

            if id_ not in anchor_ids:
                anchor_ids.add(id_)
            else:
                print(f'Duplicate anchor id "{id_}" in "{table_name}"\nduplicate data: {row_dict}\nrecord skipped')
                continue

            anchors.append(AnchorRecord(id_, type_, row_dict, dp_id=db_id))  # type: ignore

        return anchors

    def get_links(self, table_name: str, link: Link) -> list[LinkRecord]:
        table_name = f"{self.link_table_prefix}{table_name}"
        table = self.get_table(table_name)
        if "id" not in table.columns:
            table["id"] = [None] * table.shape[0]
        table_records: list[dict] = table.to_dict(orient="records")

        def flatten_list_cells(el):
            if isinstance(el, list) and "L" in el and len(el) == 2:  # flatten List type fields
                el = el[1]
            return el

        links: list[LinkRecord] = []
        for row_dict in table_records:
            id_from = row_dict.pop("from_node_id")
            id_to = row_dict.pop("to_node_id")
            # node types - check node type in id's, else take link's definition
            node_from = link.anchor_to.noun if id_from.split(":")[0] == (link.anchor_to.noun) else link.anchor_from.noun
            node_to = link.anchor_to.noun if node_from != link.anchor_to.noun else link.anchor_from.noun
            type_ = row_dict.pop("edge_label")
            row_dict.pop("id")

            if not id_from or not id_to or not type_ or pd.isna(type_) or pd.isna(id_from) or pd.isna(id_to):
                continue
            row_dict = {str(k): flatten_list_cells(v) for k, v in row_dict.items()}
            row_dict = {
                k: v for k, v in row_dict.items() if not isinstance(v, (bytes, list, type(None))) and not pd.isna(v)
            }

            links.append(LinkRecord(id_from, id_to, node_from, node_to, type_, row_dict))

        return links


class GristOfflineDataProvider(GristDataProvider):
    """
    Load from grist backup file (.grist)
    """

    def __init__(self, sqlite_path: Path) -> None:
        super().__init__()
        self._conn = sqlite3.connect(sqlite_path)

    def close(self):
        self._conn.close()

    def list_anchor_tables(self) -> list[str]:
        rows = self._conn.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' and name like '{self.anchor_table_prefix}%'"
        ).fetchall()
        return [row[0] for row in rows]

    def list_link_tables(self) -> list[str]:
        rows = self._conn.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' and name like '{self.link_table_prefix}%'"
        ).fetchall()
        return [row[0] for row in rows]

    def get_table(self, table_name: str) -> pd.DataFrame:
        cur = self._conn.execute(f"SELECT * FROM {table_name}")
        columns = [t[0] for t in cur.description]
        rows = cur.fetchall()
        return pd.DataFrame.from_records(rows, columns=columns)


class GristCsvDataProvider(GristDataProvider):
    def __init__(self, doc_id: str, grist_server: str, api_key: str | None = None) -> None:
        super().__init__()
        self.doc_id = doc_id
        self.grist_server = grist_server
        self.api_key = api_key
        self._table_list: list[str] = self._fetch_table_list()

    def _fetch_table_list(self) -> list[str]:
        url = f"{self.grist_server}/api/docs/{self.doc_id}/tables"
        resp = requests.get(url, headers={"Authorization": f"Bearer {self.api_key}"})
        resp.raise_for_status()
        return [t["id"] for t in resp.json()["tables"]]

    def get_table(self, table_name: str) -> pd.DataFrame:
        url = f"{self.grist_server}/api/docs/{self.doc_id}/download/csv?tableId={table_name}"
        resp = requests.get(url, headers={"Authorization": f"Bearer {self.api_key}"}, timeout=600)
        resp.raise_for_status()
        df = pd.read_csv(io.StringIO(resp.text))
        if "id" not in df.columns:
            df = df.reset_index(drop=False, names=["id"])
        return df

    def _list_tables_with_prefix(self, prefix: str) -> list[str]:
        return [t for t in self._table_list if t.startswith(prefix)]

    def list_anchor_tables(self) -> list[str]:
        return self._list_tables_with_prefix(self.anchor_table_prefix)

    def list_link_tables(self) -> list[str]:
        return self._list_tables_with_prefix(self.link_table_prefix)


class GristAPIDataProvider(GristDataProvider):
    def __init__(self, doc_id: str, grist_server: str, api_key: str | None = None) -> None:
        super().__init__()
        self.grist_server = grist_server
        self.doc_id = doc_id
        self.api_key = api_key
        self.headers = {"Authorization": f"Bearer {self.api_key}"}
        self._client = grist_api.GristDocAPI(doc_id, api_key=api_key, server=grist_server)

    def _list_tables_with_prefix(self, prefix: str) -> list[str]:
        resp = self._client.tables()
        if not resp:
            return []
        table_ids: list[str] = [table["id"] for table in resp.json()["tables"]]
        return [t_id for t_id in table_ids if t_id.startswith(prefix)]

    def list_anchor_tables(self) -> list[str]:
        return self._list_tables_with_prefix(self.anchor_table_prefix)

    def list_link_tables(self) -> list[str]:
        return self._list_tables_with_prefix(self.link_table_prefix)

    def _list_table_columns(self, table_name: str) -> dict[str, str]:
        """returns mapping {internal_grist_id: id_in_UI}"""
        # There is no label (hidden=True/False) on columns so we have to do 2 api calls
        view_cols = self._client.columns(table_name)  # does not show hidden columns i.e. "id"
        all_cols = requests.get(  # shows all columns, including internal IDs and helper columns
            f"{self.grist_server}/api/docs/{self.doc_id}/tables/{table_name}/columns?hidden=True", headers=self.headers
        )
        if not view_cols or not all_cols:
            return {}

        view_cols = view_cols.json()["columns"]
        all_cols = all_cols.json()["columns"]

        # get display columns
        col_ref_label_map = {c["fields"]["colRef"]: c["fields"]["label"] for c in all_cols}

        parsed_cols = {
            c["id"]: c["fields"]["label"]
            if not c["fields"]["displayCol"]
            else col_ref_label_map.get(c["fields"]["displayCol"], c["fields"]["label"])
            for c in view_cols
        }
        return parsed_cols

    def get_table(self, table_name: str) -> pd.DataFrame:
        columns = self._list_table_columns(table_name)
        if "id" not in columns:  # add internal id
            columns["id"] = "id"
        rows = self._client.fetch_table(table_name)
        rows = [{c_id: getattr(r, c_label) for c_id, c_label in columns.items()} for r in rows]  # filter usage
        return pd.DataFrame(rows, columns=list(columns.keys()))


class GristSQLDataProvider(GristDataProvider):
    """fetches rows via the /sql endpoint in chunks"""

    def __init__(
        self,
        doc_id: str,
        grist_server: str,
        api_key: str | None = None,
        *,
        batch_size: int = 800,
    ) -> None:
        super().__init__()
        self.doc_id = doc_id
        self.grist_server = grist_server
        self.api_key = api_key
        self.batch_size = max(1, batch_size)
        self.headers = {"Authorization": f"Bearer {self.api_key}"}
        self._client = grist_api.GristDocAPI(doc_id, server=grist_server, api_key=api_key)

    def reset_doc(self, doc_id: str):
        print("reopening doc")
        resp = requests.post(
            f"{self.grist_server}/api/docs/{doc_id}/force-reload",
            headers=self.headers,
            timeout=120,
        )
        resp.raise_for_status()

    def clean_doc_history(self, doc_id: str):
        print("cleaning memory")
        resp = requests.post(
            f"{self.grist_server}/api/docs/{doc_id}/states/remove",
            headers=self.headers,
            json={"keep": 1},
            timeout=120,
        )
        resp.raise_for_status()

    def _sql_endpoint(self) -> str:
        """Fully qualified URL of the /sql POST endpoint for this document."""
        return f"{self.grist_server}/api/docs/{self.doc_id}/sql"

    def _run_sql(self, sql: str, att=1) -> list[dict]:
        try:
            resp = requests.post(
                self._sql_endpoint(),
                json={"sql": sql},
                headers=self.headers,
                timeout=180,
            )
            resp.raise_for_status()
            data = resp.json()  # { "records": [ { "fields": {...} }, ... ] }.
            data = data.get("records", [])
        except Exception as e:
            print(f"Failed to fetch {sql}: {e}\n Retrying in 120 seconds...")  # todo logging
            time.sleep(120)
            self.clean_doc_history(self.doc_id)
            self.reset_doc(self.doc_id)
            if att < 3:
                att += 1
                data = self._run_sql(sql, att)
            else:
                raise e
        return data

    def _list_tables_with_prefix(self, prefix: str) -> list[str]:
        resp = self._client.tables()
        if not resp:
            return []
        table_ids: list[str] = [table["id"] for table in resp.json()["tables"]]
        return [t_id for t_id in table_ids if t_id.startswith(prefix)]

    def list_anchor_tables(self) -> list[str]:
        return self._list_tables_with_prefix(self.anchor_table_prefix)

    def list_link_tables(self) -> list[str]:
        return self._list_tables_with_prefix(self.link_table_prefix)

    def _iter_table_rows(self, table_name: str):
        offset = 0
        while True:
            sql = f'SELECT * FROM "{table_name}" ORDER BY id LIMIT {self.batch_size} OFFSET {offset}'
            batch = self._run_sql(sql)
            if not batch:
                break
            for record in batch:
                yield record.get("fields", {})
            offset += self.batch_size

    def get_anchors(self, type_: str, dm_attrs: list[Attribute], dm_anchor_links: list[Link]) -> list[AnchorRecord]:
        dtypes = {e.name: e.dtype for e in dm_attrs}
        table_name = f"{self.anchor_table_prefix}{type_}"
        default_id_key = "node_id"
        fk_links_cols = [c.anchor_from_link_attr_name for c in dm_anchor_links]

        anchors: dict[str, AnchorRecord] = {}

        def flatten(el):
            if isinstance(el, list):
                if el:
                    if el[0] == "L":
                        return el[1] if len(el) == 2 else el[1:]
            elif pd.isna(el):
                return None
            return el

        def safe_fk_list_convert(el):
            try:
                if isinstance(el, str):
                    el = literal_eval(el)
            except SyntaxError:
                print(f"str el not list: {el}")
            return el

        for row in self._iter_table_rows(table_name):
            db_id = flatten(row.get("id"))
            _id = row.get(f"{type_}_id") or row.get(default_id_key) or f"{type_}:{db_id}"
            if pd.isna(_id) or (isinstance(_id, dict) and _id.get("type") == "Buffer"):
                continue

            row.pop("id", None)
            row.pop("node_type", None)
            row.pop(default_id_key, None)
            # row.pop(f"{type_}_id", None)

            # grist meta columns
            row.pop("manualSort", None)

            row = {
                k: v
                for k, v in row.items()
                if not (isinstance(v, dict) and v.get("type") == "Buffer")
                and not k.startswith("gristHelper_")
                and not pd.isna(v)  # type: ignore
            }

            # foreign key link columns - either int id or a list of int ids cast to string
            for col in fk_links_cols:
                if col in row:
                    row[col] = safe_fk_list_convert(row[col])

            clean_data = {
                k: cast_dtype(flatten(v), k, dtypes.get(k))
                for k, v in row.items()
                if not isinstance(v, (bytes, type(None))) and v != ""
            }

            if _id in anchors:  # node already present - populate its attributes
                # anchors[_id].data.update(clean_data)
                print(f"duplicate {type_} id {_id}, skipping...")
            else:
                anchors[_id] = AnchorRecord(str(_id), type_, clean_data, dp_id=db_id)  # type: ignore

        return list(anchors.values())

    def get_links(self, table_name: str, link: Link) -> list[LinkRecord]:
        table_name = f"{self.link_table_prefix}{table_name}"
        columns_resp = self._client.columns(table_name)
        if not columns_resp:
            return []

        def flatten(el):
            if isinstance(el, list) and "L" in el and len(el) == 2:
                el = el[1]
            return el

        links: list[LinkRecord] = []
        for row in self._iter_table_rows(table_name):
            id_from = row.get("id_from") or row.get("from_node_id")
            id_to = row.get("id_to") or row.get("to_node_id")
            edge_label = row.get("type") or row.get("edge_label")
            if not id_from or not id_to or not edge_label or pd.isna(edge_label) or pd.isna(id_from) or pd.isna(id_to):
                continue

            row.pop("manualSort", None)
            row.pop("from_node_id", None)
            row.pop("to_node_id", None)
            row.pop("edge_label", None)
            row.pop("id_from", None)
            row.pop("id_to", None)
            row.pop("type", None)
            row.pop("id", None)

            clean_data = {
                k: flatten(v)
                for k, v in row.items()
                if not isinstance(v, (bytes, list))
                and not pd.isna(v)
                and not (isinstance(v, str) and v == "")
                and not k.startswith("gristHelper_")
            }

            links.append(
                LinkRecord(
                    str(id_from),
                    str(id_to),
                    link.anchor_from.noun,
                    link.anchor_to.noun,
                    str(edge_label),
                    clean_data,
                )
            )

        return links

    def get_table(self, table_name: str) -> pd.DataFrame:
        """
        Table with 1st batch_size rows.
        To iterate over the remaining rows use _iter_table_rows
        """
        cols_resp = self._client.columns(table_name)
        columns: list[str] = []
        if cols_resp:
            columns = [c["id"] for c in cols_resp.json()["columns"]]

        rows_data = []
        for i, row in enumerate(self._iter_table_rows(table_name)):
            if i >= self.batch_size:
                break
            # row dict to tuple in deterministic column order
            rows_data.append(tuple(row.get(col) for col in columns))

        return pd.DataFrame(rows_data, columns=columns)



import abc
import json
import logging
import re
from typing import Any, Dict, Iterable, Set, cast

import aioitertools as aioit
import neo4j
import numpy as np
import typing_extensions as te
from neo4j import AsyncGraphDatabase, EagerResult, RoutingControl
from opentelemetry import trace

logger = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)

Record = neo4j.Record


class Graph(abc.ABC):
    async def add_node(
        self,
        node_id: str,
        labels: Set[str],
        properties: dict[str, Any] | None = None,
        embeddings: dict[str, np.ndarray] | None = None,
    ) -> None:
        raise NotImplementedError

    async def add_edge(self, from_id: str, to_id: str, type_: str, attrs: Dict[str, Any] | None) -> None:
        raise NotImplementedError

    async def number_of_nodes(self) -> int:
        raise NotImplementedError

    async def number_of_edges(self) -> int:
        raise NotImplementedError

    async def run_cypher(
        self,
        query: str,
        parameters: dict[str, Any] | None = None,
        limit: int | None = None,
    ) -> Iterable[Record]:
        raise NotImplementedError

    async def get_existing_node_types(self) -> Iterable[list[str]]:
        raise NotImplementedError

    async def llm_schema(self) -> str:
        raise NotImplementedError

    async def text_search(self, label: str, query: str, limit: int = 10) -> Iterable[Record]:
        raise NotImplementedError

    async def setup(self, *_, create_basic_indices: bool = True, **kwargs) -> None:
        # Set false to speedup import
        if create_basic_indices:
            await self.create_basic_indices()

    async def create_basic_indices(self) -> None:
        ...

    async def execute_ro_cypher_query(
        self,
        query: str,
        parameters: dict[str, Any] | None = None,
        limit: int | None = None,
    ) -> Iterable[Record]:
        return await self.run_cypher(query, parameters, limit=limit)

    async def clear(self) -> None:
        ...

    def close(self) -> None:
        ...

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class CypherGraph(Graph):
    async def add_node(
        self,
        node_id: str,
        labels: Set[str],
        properties: dict[str, Any] | None = None,
        embeddings: dict[str, np.ndarray] | None = None,
    ) -> None:
        query, params = self._add_node_cypher(node_id, labels, properties or {})
        await self.run_cypher(query, params)

    def _add_node_cypher(
        self,
        node_id: str,
        labels: Set[str],
        properties: dict[str, Any],
    ) -> tuple[str, dict[str, Any]]:
        labels_expr = escape_labels(labels)
        props = {
            **properties,
            "id": node_id,
        }
        # TODO escape
        pros_expr = ", ".join(f"{k}: ${k}" for k in props.keys())
        return (
            f"MERGE (n:{labels_expr} {{id: $id}}) SET n = {{{pros_expr}}} RETURN n",
            props,
        )

    async def add_edge(self, from_id: str, to_id: str, type_: str, attrs: Dict[str, Any] | None) -> None:
        query, params = self._add_edge_cypher(from_id, to_id, type_, attrs)
        await self.run_cypher(query, params)

    def _add_edge_cypher(
        self, from_id: str, to_id: str, type_: str, attrs: Dict[str, Any] | None
    ) -> tuple[str, dict[str, Any]]:
        attrs = attrs or {}
        labels_expr = escape_labels({type_})
        # attrs = {escape_cypher(k): v for k, v in attrs.items()}
        attrs_expr = ", ".join(f"{k}: ${k}" for k in attrs.keys() if k)
        params = {
            **attrs,
            "from_id": from_id,
            "to_id": to_id,
        }
        return (
            "MATCH (nf {id: $from_id}), (nt {id: $to_id}) "
            f"CREATE (nf)-[r:{labels_expr} {{{attrs_expr}}}]->(nt) RETURN r",
            params,
        )

    async def add_edges(self, edges: Iterable[tuple[str, str, dict]], **common_attrs) -> None:
        for edge_tuple in edges:
            from_id, to_id, attrs = edge_tuple
            attrs = {**common_attrs, **attrs}
            labels: Iterable[str] = attrs.pop("__labels__", [])
            type_ = next(iter(labels), "no_type")
            await self.add_edge(from_id, to_id, type_, attrs)

    async def number_of_nodes(self) -> int:
        res = await self.execute_ro_cypher_query("MATCH (n) RETURN count(*) as cnt")
        return next(iter(res))["cnt"]

    async def number_of_edges(self) -> int:
        res = await self.execute_ro_cypher_query("MATCH (f)-[]->(t) RETURN count(*) as cnt")
        return next(iter(res))["cnt"]

    async def get_existing_node_types(self) -> Iterable[list[str]]:
        res = await self.execute_ro_cypher_query("MATCH (n) RETURN DISTINCT labels(n) as l;")
        return [r["l"] for r in res]


# class NXGraph(Graph):
#     def __init__(self, graph: nx.Graph) -> None:
#         self.graph: nx.Graph = graph
#         self.gcypher = GrandCypher(self.graph)

#     def execute_ro_cypher_query(self, query: str) -> Iterable[Any]:
#         return self.gcypher.run(query)

#     def add_node(self, node_id: str, labels: Set[str], **attributes) -> None:
#         self.graph.add_node(node_id, __labels__=labels, **attributes)

#     def number_of_edges(self) -> int:
#         return self.graph.number_of_edges()

#     def clear(self) -> None:
#         self.graph.clear()


class MemgraphGraph(CypherGraph):
    def __init__(self, uri: str, user: str, pwd: str, db_name: str = "") -> None:
        self.driver = AsyncGraphDatabase.driver(uri, auth=(user, pwd), database=db_name)
        # await self.driver.verify_connectivity()
        self.driver_uri = uri
        self.auth = (user, pwd)

    async def execute_ro_cypher_query(
        self, query: str, parameters: dict[str, Any] | None = None, limit: int | None = None
    ) -> Iterable[Record]:
        with tracer.start_as_current_span("memgraph.execute_ro_cypher_query") as span:
            span.set_attribute("memgraph.query", query)
            if parameters:
                span.set_attribute("memgraph.parameters", json.dumps(parameters))
            result: EagerResult = await self.driver.execute_query(query, parameters, routing_=RoutingControl.READ)

        return result.records

    async def run_cypher(
        self,
        query: str,
        parameters: dict[str, Any] | None = None,
        limit: int | None = None,
    ) -> Iterable[Record]:
        with tracer.start_as_current_span("memgraph.run_cypher") as span:
            span.set_attribute("memgraph.query", query)
            if parameters:
                span.set_attribute("memgraph.parameters", json.dumps(parameters))
            if limit is not None:
                span.set_attribute("memgraph.limit", limit)

            async with self.driver.session() as session:
                result = await aioit.list(aioit.islice(await session.run(query, parameters), limit))

        return result

    async def add_node(
        self,
        node_id: str,
        labels: Set[str],
        properties: Dict[str, Any] | None = None,
        embeddings: Dict[str, np.ndarray] | None = None,
    ) -> None:
        if properties and embeddings:
            embed_props = {f"{prop_name}_embedding": v for prop_name, v in embeddings.items()}
            properties = {
                **properties,
                **embed_props,
            }
        await super().add_node(node_id, labels, properties, embeddings)

    async def llm_schema(self) -> str:
        """can be used as fallback data model structure"""
        res = await self.driver.execute_query("CALL llm_util.schema() YIELD schema RETURN schema")
        return res.records[0]["schema"]

    async def create_basic_indices(self, node_types=None) -> None:
        if not node_types:
            node_types = await self.get_existing_node_types()
        for label in node_types:
            await self.create_node_prop_index(set(label), "id", unique=True)

    async def clear(self) -> None:
        async with self.driver.session() as session:
            res = await session.run("CALL vector_search.show_index_info() YIELD index_name RETURN *")

            async for (idx_name,) in res:
                await session.run(f"DROP VECTOR INDEX {escape_cypher(idx_name)}")
            idx_name_re = re.compile(r"\(name:\s(.+?)\)")
            async for row in await session.run(cast(te.LiteralString, "SHOW INDEX INFO")):
                index_type = row["index type"]
                idx_name = next(iter(idx_name_re.findall(index_type)), None)
                if not idx_name:
                    continue
                await session.run(f"DROP TEXT INDEX {escape_cypher(idx_name)}")
            await session.run("CALL schema.assert({}, {}, {}, true) YIELD action, key, keys, label, unique")
            await session.run("MATCH (n) DETACH DELETE n")
            # TODO more efficient:
            # USING PERIODIC COMMIT num_rows
            # MATCH (n)-[r]->(m)
            # DELETE r;
            # USING PERIODIC COMMIT num_rows
            # MATCH (n)
            # DETACH DELETE n;

    async def text_search(self, label: str, query: str, limit: int = 10) -> Iterable[Record]:
        with tracer.start_as_current_span("memgraph.text_search") as span:
            span.set_attribute("memgraph.label", label)
            span.set_attribute("memgraph.fts_query", query)
            span.set_attribute("memgraph.limit", limit)

            query = "CALL text_search.search_all($idx_name, $query) YIELD node RETURN node LIMIT $limit"
            span.set_attribute("memgraph.query", query)

            res = await self.driver.execute_query(
                query,
                idx_name=self._fts_idx_name(label),
                query=query,
                limit=limit,
                routing_=RoutingControl.READ,
            )
            return res.records

    async def create_node_prop_index(self, labels: set[str], property: str, unique: bool = False) -> None:
        escaped_label = escape_labels(labels)
        escaped_prop = escape_cypher(property)
        await self.run_cypher(f"CREATE INDEX ON :{escaped_label}({escaped_prop})")
        if not unique:
            return
        await self.run_cypher(f"CREATE CONSTRAINT ON (n:{escaped_label})\nASSERT n.{escaped_prop} IS UNIQUE")

    @staticmethod
    def _fts_idx_name(label: str) -> str:
        return f"{label.lower()}_fts_idx"

    def close(self):
        self.driver.close()


def escape_cypher(identifier: str) -> str:
    identifier = identifier.replace("\u0060", "`").replace("`", "``")
    return f"`{identifier}`"


def escape_labels(labels: set[str]) -> str:
    return ":".join(escape_cypher(label) for label in labels)



import abc
import logging
from typing import Sequence

import numpy as np
from neo4j import AsyncGraphDatabase, Record, RoutingControl
from opentelemetry import trace
from sqlalchemy import RowMapping, select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker
from vedana_etl.catalog import edges, nodes, rag_anchor_embeddings, rag_edge_embeddings

from vedana_core.db import get_sessionmaker

logger = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)


class VectorStore(abc.ABC):
    async def vector_search(
        self,
        label: str,
        prop_type: str,
        prop_name: str,
        embedding: list[float],
        threshold: float,
        top_n: int = 5,
    ):
        raise NotImplementedError


class MemgraphVectorStore(VectorStore):
    """
    Use Memgraph vector_search capabilities. Requires vector indices to be created separately.
    """

    def __init__(self, uri: str, user: str, pwd: str, db_name: str = "") -> None:
        self.driver = AsyncGraphDatabase.driver(uri, auth=(user, pwd), database=db_name)
        # await self.driver.verify_connectivity()
        self.driver_uri = uri
        self.auth = (user, pwd)

    async def vector_search(
        self,
        label: str,
        prop_type: str,
        prop_name: str,
        embedding: np.ndarray | list[float],
        threshold: float,
        top_n: int = 5,
    ) -> list[Record]:
        with tracer.start_as_current_span("memgraph.vector_search") as span:
            span.set_attribute("memgraph.label", label)
            span.set_attribute("memgraph.prop_type", prop_type)
            span.set_attribute("memgraph.prop_name", prop_name)
            span.set_attribute("memgraph.top_n", top_n)
            span.set_attribute("memgraph.threshold", threshold)

            if prop_type == "edge":
                query = (
                    "CALL vector_search.search_edges($idx_name, $top_n, $embedding) "
                    "YIELD similarity, edge "
                    "WITH similarity, edge "
                    "WHERE similarity > $threshold "
                    "RETURN similarity, edge, startNode(edge) AS start, endNode(edge) AS end;"
                )
            else:  # node
                query = (
                    "CALL vector_search.search($idx_name, $top_n, $embedding) "
                    "YIELD similarity, node "
                    "WITH similarity, node WHERE similarity > $threshold RETURN *"
                )

            span.set_attribute("memgraph.query", query)

            idx_name = f"{label}_{prop_name}_embed_idx".replace(" ", "_")
            res = await self.driver.execute_query(
                query,
                idx_name=idx_name,
                top_n=top_n,
                embedding=embedding,
                threshold=threshold,
                routing_=RoutingControl.READ,
            )
            return res.records


class PGVectorStore(VectorStore):
    def __init__(
        self,
        sessionmaker: async_sessionmaker[AsyncSession] | None = None,
    ) -> None:
        self._sessionmaker: async_sessionmaker[AsyncSession] = sessionmaker or get_sessionmaker()
        self.rag_anchor_embeddings_table = rag_anchor_embeddings.store.data_table  # type: ignore[attr-defined]
        self.rag_edge_embeddings_table = rag_edge_embeddings.store.data_table  # type: ignore[attr-defined]
        self.node_table = nodes.store.data_table  # type: ignore[attr-defined]
        self.edge_table = edges.store.data_table  # type: ignore[attr-defined]

    async def vector_search(
        self,
        label: str,
        prop_type: str,
        prop_name: str,
        embedding: np.ndarray | list[float],
        threshold: float,
        top_n: int = 5,
    ) -> Sequence[RowMapping]:
        with tracer.start_as_current_span("pgvector.vector_search") as span:
            span.set_attribute("pgvector.label", label)
            span.set_attribute("pgvector.prop_type", prop_type)
            span.set_attribute("pgvector.prop_name", prop_name)
            span.set_attribute("pgvector.top_n", top_n)
            span.set_attribute("pgvector.threshold", threshold)

            async with self._sessionmaker() as session:
                if prop_type == "edge":
                    distance = self.rag_edge_embeddings_table.c.embedding.cosine_distance(embedding)
                    similarity = (1 - distance).label("similarity")

                    stmt = (
                        select(
                            similarity,
                            self.edge_table.c.from_node_id,
                            self.edge_table.c.to_node_id,
                            self.edge_table.c.edge_label,
                            self.edge_table.c.attributes.label("edge"),
                        )
                        .select_from(
                            self.rag_edge_embeddings_table.join(
                                self.edge_table,
                                (self.rag_edge_embeddings_table.c.from_node_id == self.edge_table.c.from_node_id)
                                & (self.rag_edge_embeddings_table.c.to_node_id == self.edge_table.c.to_node_id)
                                & (self.rag_edge_embeddings_table.c.edge_label == self.edge_table.c.edge_label),
                            )
                        )
                        .where(self.rag_edge_embeddings_table.c.edge_label == label)
                        .where(self.rag_edge_embeddings_table.c.attribute_name == prop_name)
                        .where(similarity > threshold)
                        .order_by(distance)
                        .limit(top_n)
                    )

                else:  # node
                    distance = self.rag_anchor_embeddings_table.c.embedding.cosine_distance(embedding)
                    similarity = (1 - distance).label("similarity")

                    stmt = (
                        select(
                            similarity,
                            self.node_table.c.node_id,
                            self.node_table.c.node_type,
                            self.node_table.c.attributes.label("node"),
                        )
                        .select_from(
                            self.rag_anchor_embeddings_table.join(
                                self.node_table, self.rag_anchor_embeddings_table.c.node_id == self.node_table.c.node_id
                            )
                        )
                        .where(self.rag_anchor_embeddings_table.c.label == label)
                        .where(self.rag_anchor_embeddings_table.c.attribute_name == prop_name)
                        .where(similarity > threshold)
                        .order_by(distance)
                        .limit(top_n)
                    )

                span.set_attribute("pgvector.query", str(stmt))
                res = await session.execute(stmt)
                return res.mappings().all()



from dataclasses import dataclass

import sqlalchemy.ext.asyncio as sa_aio
from async_lru import alru_cache
from jims_core.app import JimsApp
from loguru import logger

from vedana_core.data_model import DataModel
from vedana_core.db import get_sessionmaker
from vedana_core.graph import Graph, MemgraphGraph
from vedana_core.vts import VectorStore, PGVectorStore
from vedana_core.rag_pipeline import RagPipeline
from vedana_core.start_pipeline import StartPipeline
from vedana_core.settings import settings as core_settings


@dataclass
class VedanaApp:
    sessionmaker: sa_aio.async_sessionmaker[sa_aio.AsyncSession]

    graph: Graph
    vts: VectorStore
    data_model: DataModel
    pipeline: RagPipeline
    start_pipeline: StartPipeline


@alru_cache
async def make_vedana_app() -> VedanaApp:
    sessionmaker = get_sessionmaker()

    graph = MemgraphGraph(
        core_settings.memgraph_uri,
        core_settings.memgraph_user,
        core_settings.memgraph_pwd,
    )

    vts = PGVectorStore(
        sessionmaker=sessionmaker,
    )

    data_model = DataModel(sessionmaker=sessionmaker)

    pipeline = RagPipeline(
        graph=graph,
        vts=vts,
        data_model=data_model,
        logger=logger,
        threshold=0.8,
    )

    start_pipeline = StartPipeline(data_model=data_model)

    return VedanaApp(
        sessionmaker=sessionmaker,
        graph=graph,
        vts=vts,
        data_model=data_model,
        pipeline=pipeline,
        start_pipeline=start_pipeline,
    )


@alru_cache
async def make_jims_app() -> JimsApp:
    vedana_app = await make_vedana_app()

    app = JimsApp(
        sessionmaker=vedana_app.sessionmaker,
        pipeline=vedana_app.pipeline,
        conversation_start_pipeline=vedana_app.start_pipeline,
    )

    return app


# This creates a async coroutine which will be evaluated in the event loop
app = make_jims_app()



from pydantic_settings import BaseSettings, SettingsConfigDict
from dotenv import load_dotenv

load_dotenv()


class VedanaCoreSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="",
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )

    grist_server_url: str  # default 'https://api.getgrist.com'
    grist_api_key: str
    grist_data_model_doc_id: str
    grist_data_doc_id: str

    debug: bool = False
    model: str = "gpt-4.1"
    enable_dm_filtering: bool = False  # pipeline selection (experimental)
    filter_model: str = "gpt-5-mini"
    judge_model: str = "gpt-5-mini"
    embeddings_model: str = "text-embedding-3-large"
    embeddings_dim: int = 1024

    pipeline_history_length: int = 20

    memgraph_uri: str
    memgraph_user: str
    memgraph_pwd: str


settings = VedanaCoreSettings()  # type: ignore
